"""
This module defines adds some common types to the default :mod:`typing` module present in the standard library.

It is recommended to import it with *one* of the following statements::

    import royalnet.royaltyping as t
    from royalnet.royaltyping import *
    from royalnet.royaltyping import <used_objects>

"""

from __future__ import annotations
from typing import *
# noinspection PyUnresolvedReferences
from typing import IO, TextIO, BinaryIO
# noinspection PyUnresolvedReferences
from typing import Pattern, Match


JSONScalar = Union[
    None,
    float,
    int,
    str,
]
"""
A non-recursive JSON value: either :data:`None`, a :class:`float`, a :class:`int` or a :class:`str`.
"""

JSON = Union[
    JSONScalar,
    List["JSON"],
    Dict[str, "JSON"],
]
"""
A recursive JSON value: either a :data:`.JSONScalar`, or a :class:`list` of :data:`.JSON` objects, or a :class:`dict` 
of :class:`str` to :data:`.JSON` mappings. 
"""


WrenchLike = Callable[[Any], Awaitable[Any]]


class ConversationProtocol(Protocol):
    def __call__(self, **kwargs) -> Awaitable[None]:
        ...


Args = Collection[Any]
"""
Any possible combination of positional arguments.
"""

Kwargs = Mapping[str, Any]
"""
Any possible combination of keyword arguments.
"""



"""
This module exports the base exceptions used in all :mod:`royalnet` modules.
"""


class RoyalnetException(Exception):
    """An exception raised by a Royalnet module."""


__all__ = ("RoyalnetException",)



"""
Royalnet is a collection of many Python modules useful to build chat-like applications.

The modules currently are:

- :mod:`.alchemist`, containing utility extensions to :mod:`sqlalchemy`;
- :mod:`.engineer`, containing a framework for building chat bots;
- :mod:`.lazy`, containing utilities to delay the evaluation of values until they are actually used;
- :mod:`.scrolls`, containing configuration utilities;
- :mod:`.sculptor`, containing common :mod:`pydantic` models for serialization and deserialization of data structures;

To prevent the library from breaking if optional dependencies are not installed, this module does not export any object; submodules should be directly imported instead.

All modules use exceptions based on :exc:`.exc.RoyalnetException`, and may subclass it to provide more detail on the errors.
"""
